import aiounittest


from amqtt_db.plugin import DBPlugin
from amqtt_db.tests.resources.mockups import get_context, CONFIG_OK, CONFIG_AMQTT_DB_MISSING, \
    CONFIG_PLUGIN_NAME_WRONG, CONFIG_TOPIC_ENGINE_WRONG, CONFIG_NO_DB_CONNECT_STRING, CONFIG_NO_DB_STRUCTURE, \
    CONFIG_WRONG_DB_STRUCTURE


class TestPlugin(aiounittest.AsyncTestCase):

    def test_plugin_loads(self):
        context = get_context(CONFIG_OK)
        _plugin = DBPlugin(context)    # noqa: F841

    def test_plugin_wrong_config(self):
        context = get_context(CONFIG_PLUGIN_NAME_WRONG)
        self.assertRaises(ImportError, DBPlugin, context)

    def test_plugin_amqtt_section_missing(self):
        context = get_context(CONFIG_AMQTT_DB_MISSING)
        self.assertRaises(ImportError, DBPlugin, context)

    def test_plugin_topic_engine_wrong(self):
        context = get_context(CONFIG_TOPIC_ENGINE_WRONG)
        self.assertRaises(ImportError, DBPlugin, context)

    def test_plugin_no_db_connect_string(self):
        context = get_context(CONFIG_NO_DB_CONNECT_STRING)
        self.assertRaises(ImportError, DBPlugin, context)

    def test_plugin_no_db_structure(self):
        context = get_context(CONFIG_NO_DB_STRUCTURE)
        self.assertRaises(ImportError, DBPlugin, context)

    def test_plugin_wrong_db_structure(self):
        context = get_context(CONFIG_WRONG_DB_STRUCTURE)
        self.assertRaises(ImportError, DBPlugin, context)



import weakref


class BaseDB(object):
    """
    Base API to the data base
    """

    def __init__(self, context, connect_string, **kwargs):
        self._context = weakref.ref(context)
        self.logger = self.context.logger
        self.init_db(connect_string, **kwargs)

    @property
    def context(self):
        return self._context()

    def init_db(self, connect_string, **kwargs):
        """
        Initialize the DB
        """

    def create_session_table(self):
        """
        Creates the session table
        """

    def create_table(self, name, column_def):
        """
        Creates a table based on table_name and column definition. Where Column_def is a dict of col_name, col_type
        """

    def add_new_columns(self, table, columns):
        """
        Extends a table with some columns
        """

    def drop_table(self):
        """
        Creates a table
        """

    def empty_table(self):
        """
        Empties a table
        """

    def add_packet(self, session, sender, topic, data):
        """
        persist a packet
        """



from pydoc import locate

from amqtt_db.base.errors import TypeNotFound


def get_class_func_by_name(class_func_path):
    """
    This simple one liner is a security hazard. Here takes place the conversion of config strings to living
    python classes or functions.
    If your config files are not protected this is the place the hacker will attack.
    :param class_func_path: A given full qualified class or function locator.
    :return: A python object of this thing
    """
    cls = locate(class_func_path)
    if cls is None:
        msg = 'While doing a string to type conversion the type: "{}" cannot be found.'.format(class_func_path)
        raise TypeNotFound(msg)

    return cls



"""
Base class for payload decoders
"""


class BaseDecoder(object):

    @staticmethod
    def decode(payload):
        """
        Gecodes a given payload
        :param payload: the payload
        :return: the decoded payload a python data structures
        """



import os
from datetime import date, time, datetime

import unittest

from amqtt_db.db.db_sa import SA
from amqtt_db.tests.resources.mockups import get_context, CONFIG_OK

from tempfile import TemporaryDirectory


class TestSA(unittest.TestCase):

    def create_test_db(self):
        temp_dir = TemporaryDirectory()
        temp_db_file_name = os.path.join(temp_dir.name, 'huhu')
        context = get_context(CONFIG_OK)
        sa = SA(context, 'sqlite:///{}'.format(temp_db_file_name), echo=True)
        return temp_dir, sa

    def test_create_table(self):
        tempdir, sa = self.create_test_db()

        column_def = {
            'a_name': type('ABC'),
            'a_date': type(date(2021, 1, 2)),
            'a_time': type(time(23, 34, 12)),
            'a_datetime': type(datetime(2021, 1, 2, 23, 34, 12)),
            'a_int': type(123),
            'a_float': type(123.456),
            }

        sa.create_table('tests', column_def)
        tempdir.cleanup()

    def test_add_packet(self):
        tempdir, sa = self.create_test_db()

        data = {
            'a_name': 'Hugo der Erste der Überflieger',
            'a_date': date(2021, 1, 2),
            'a_time': time(23, 34, 12),
            'a_datetime': datetime(2021, 1, 2, 23, 34, 12),
            'a_int': 1212334,
            'a_float': 0.1234,
            }

        sa.add_packet('test1', 'abcdef', 'test/topic', data)

        query = sa.query_topic('test/topic', data)
        query_result = query.all()
        assert len(query_result) == 1
        assert query_result != [None]
        tempdir.cleanup()

    def test_packet_growth(self):
        tempdir, sa = self.create_test_db()

        data1 = {
            'a_name': 'Hugo der Erste der Überflieger',
            'a_date': date(2021, 1, 2),
            'a_time': time(23, 34, 12),
        }

        data2 = {
            'a_name': 'Otto der Zweite der Überflieger',
            'a_date': date(2021, 1, 2),
            'a_time': time(23, 34, 2),
            'a_datetime': datetime(2021, 1, 2, 23, 34, 12),
            'a_int': 1212334,
            'a_float': 0.1234,
            }

        sa.add_packet('test1', 'abcdef', 'test/topic_growth', data1)

        query = sa.query_topic('test/topic_growth', data1)
        query_result = query.all()
        assert len(query_result) == 1
        assert query_result != [None]

        sa.add_packet('test1', 'abcdef', 'test/topic_growth', data2)

        tempdir.cleanup()



"""
Shamelessly stolen from
https://www.michaelcho.me/article/method-delegation-in-python
"""


class Delegator(object):
    def __getattr__(self, called_method):
        def __raise_standard_exception():
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, called_method))

        def wrapper(*args, **kwargs):
            delegation_config = getattr(self, 'DELEGATED_METHODS', None)
            if not isinstance(delegation_config, dict):
                __raise_standard_exception()

            delegate_object_str = ''
            for delegate_object_str, delegated_methods in delegation_config.items():
                if called_method in delegated_methods:
                    break
            else:
                __raise_standard_exception()

            delegate_object = getattr(self, delegate_object_str, None)

            return getattr(delegate_object, called_method)(*args, **kwargs)

        return wrapper



import aiounittest

from amqtt_db.plugin import DBPlugin
from amqtt_db.tests.resources.mockups import publish_packet, get_context, CONFIG_OK, CONFIG_TOPIC_WILDCARD, CONFIG_MISSING_RH_MEASURE


class TestPlugin2(aiounittest.AsyncTestCase):

    async def test_on_packet(self):
        context = get_context(CONFIG_OK)
        plugin = DBPlugin(context)
        await plugin.on_mqtt_packet_received(packet=publish_packet)


class TestPlugin3(aiounittest.AsyncTestCase):

    async def test_wildcard_topic(self):
        context = get_context(CONFIG_TOPIC_WILDCARD)
        plugin = DBPlugin(context)
        await plugin.on_mqtt_packet_received(packet=publish_packet)


class TestPlugin4(aiounittest.AsyncTestCase):

    async def test_initial_package_with_missing_payload_measure_entry(self):
        context = get_context(CONFIG_MISSING_RH_MEASURE)
        plugin = DBPlugin(context)
        await plugin.on_mqtt_packet_received(packet=publish_packet)



from unittest.mock import MagicMock

import yaml
from hbmqtt.broker import BrokerContext
from hbmqtt.mqtt import PublishPacket
from hbmqtt.mqtt.publish import PublishPayload, PublishVariableHeader

import logging


CONFIG_OK = """
plugins:
  - amqtt_db

amqtt_db:
  db_structure : amqtt_db.structure.structure.WideStructure
  db_connection : 'sqlite:///memory'

  payload :
    hall/temp_rh :
      amqtt_db.payload.json_decoder.JSONDecoder :
        amqtt_db.payload.flat_deserializer.FlatDeserializer :
          temp : float
          rh : float
          press : float
          gas : float
          'current time' : datetime.datetime.utcfromtimestamp
"""

CONFIG_PLUGIN_NAME_WRONG = """
plugins:
  - amqtt_db22
"""

CONFIG_AMQTT_DB_MISSING = """
plugins:
  - amqtt_db
"""

CONFIG_TOPIC_ENGINE_WRONG = """
plugins:
  - amqtt_db

amqtt_db:
    topic_engine : None
"""

CONFIG_NO_DB_CONNECT_STRING = """
plugins:
  - amqtt_db

amqtt_db:
    payload :
        hall/temp_rh :
          amqtt_db.payload.json_decoder.JSONDecoder :
            amqtt_db.payload.flat_deserializer.FlatDeserializer :
              temp : float
"""

CONFIG_NO_DB_STRUCTURE = """
plugins:
  - amqtt_db

amqtt_db:
    db_connection : 'sqlite:///memory'
    payload :
        hall/temp_rh :
          amqtt_db.payload.json_decoder.JSONDecoder :
            amqtt_db.payload.flat_deserializer.FlatDeserializer :
              temp : float
"""


CONFIG_WRONG_DB_STRUCTURE = """
plugins:
  - amqtt_db

amqtt_db:
    db_structure : amqtt_db.structure.structure.WideStru
    db_connection : 'sqlite:///memory'
    payload :
        hall/temp_rh :
          amqtt_db.payload.json_decoder.JSONDecoder :
            amqtt_db.payload.flat_deserializer.FlatDeserializer :
              temp : float
"""

CONFIG_TOPIC_WILDCARD = """
plugins:
  - amqtt_db

amqtt_db:
    db_structure : amqtt_db.structure.structure.WideStructure
    db_connection : 'sqlite:///memory'
    payload :
        hall/* :
          amqtt_db.payload.json_decoder.JSONDecoder :
            amqtt_db.payload.flat_deserializer.FlatDeserializer :
              temp : float
              press : float
              gas : float
              rh : float
              'current time' : datetime.datetime.utcfromtimestamp
"""

CONFIG_MISSING_RH_MEASURE = """
plugins:
  - amqtt_db

amqtt_db:
    db_structure : amqtt_db.structure.structure.WideStructure
    db_connection : 'sqlite:///memory'
    payload :
        hall/temp_rh :
          amqtt_db.payload.json_decoder.JSONDecoder :
            amqtt_db.payload.flat_deserializer.FlatDeserializer :
              temp : float
              press : float
              gas : float
              'current time' : datetime.datetime.utcfromtimestamp
"""


def get_context(config):
    context = MagicMock(spec=BrokerContext)()
    context.config = yaml.load(config, Loader=yaml.FullLoader)
    context.logger = logging.getLogger('test')
    context.logger.handlers = [logging.NullHandler()]
    return context


publish_packet = PublishPacket()
publish_packet.variable_header = PublishVariableHeader('hall/temp_rh')

payload = PublishPayload()
payload.data = bytearray(b'''{"3c610515b7c4": {"current time": 1617663897, "gas": 46.922, "temp": 24.35012, "press": 936.7473, "rh": 29.88471}}''')

publish_packet.payload = payload



"""
Base of a topic engine, routing the payload to the configured deserializer
"""
import re

from amqtt_db.base.constants import PAYLOAD_CONFIG
from amqtt_db.base.errors import TopicNotFound, NoPayloadDefinition, WrongPayloadDefinition
from amqtt_db.base.utils import get_class_func_by_name


class BaseTopicEngine(object):

    def __init__(self, parent):
        self.parent = parent
        self.topic_handlers = {}    # topic to handler mapping
        self.decoders = {}
        self.deserializers = {}
        self.topic_re = {}    # topic to handler mapping
        self.read_config(parent.config)

    def read_config(self, config):
        """
        Build a simple parse tree out of the payload config
        :param config: The payload config
        """
        try:
            payload_config = config[PAYLOAD_CONFIG]
        except KeyError:
            raise NoPayloadDefinition
        try:
            for key, value in payload_config.items():
                entry = list(value.items())[0]
                decoder_cls_name = entry[0]
                decoder_class = get_class_func_by_name(decoder_cls_name)
                if decoder_cls_name in self.decoders:
                    decoder = self.decoders[decoder_cls_name]
                else:
                    decoder = decoder_class()
                    self.decoders[decoder_cls_name] = decoder

                deserialier_config = entry[1]
                deserializer_cls_name, types = list(deserialier_config.items())[0]
                deserializer_class = get_class_func_by_name(deserializer_cls_name)

                if deserializer_cls_name in self.deserializers:
                    deserializer = self.deserializers[deserializer_cls_name]
                else:
                    deserializer = deserializer_class(types)
                    self.deserializers[deserializer_cls_name] = deserializer

                self.topic_handlers[key] = [decoder, deserializer]
                self.topic_re[re.compile(key)] = self.topic_handlers[key]
        except (KeyError, AttributeError):
            raise WrongPayloadDefinition

    def topic2handler(self, topic):
        """
        Get a handler for a topic according to the payload definitions
        :param topic:
        :return: The handler
        """
        if topic in self.topic_handlers:
            return self.topic_handlers[topic]
        else:
            result = self.match_topic(topic)
            if result is not None:
                self.topic_handlers[topic] = result
                return result
        raise TopicNotFound()

    def match_topic(self, topic):
        """
        Match a topic to an handler according to the config
        :param topic:
        :return:
        """
        for regex, handler in self.topic_re.items():
            if regex.match(topic):
                self.topic_handlers[topic] = self.topic_re[regex]
                return self.topic_re[regex]

        return None



from hbmqtt.mqtt import PublishPacket

from amqtt_db.base.base_structure import BaseStructure


class WideStructure(BaseStructure):
    """
    Mapping each MQTT topics to a table. Any content in the Payload is a column in the table.
    """

    # async def on_save_session(self, session):
    #     """
    #     Persist the session
    #     :param session: The current session
    #     """
    #     self.logger.debug("saving session {}".format(session))

    async def on_mqtt_packet_received(self, packet=None, session=None):
        if not isinstance(packet, PublishPacket):
            return

        self.logger.debug("saving package {}".format(packet))
        decoder, deserializer = self.topic_engine.topic2handler(packet.topic_name)
        topic = packet.topic_name

        decoded_payload = decoder.decode(packet.payload.data)

        sender, data = list(decoded_payload.items())[0]
        typed_data, residual_data = deserializer.deserialize(data)  # ToDo: Handle residual data
        typed_data['sender'] = int(sender, 16)
        self.db.add_packet(session, sender, self.topic2SQL(topic), typed_data)

        self.logger.debug("package {} saved".format(packet))



import weakref


class BaseStructure(object):
    """
    API for the mqtt handlers
    """

    def __init__(self, topic_engine, db, context):
        self._topic_engine = weakref.ref(topic_engine)
        self._context = weakref.ref(context)
        self._db = weakref.ref(db)
        self.logger = self.context.logger

    @property
    def db(self):
        return self._db()

    @property
    def context(self):
        return self._context()

    @property
    def topic_engine(self):
        return self._topic_engine()

    def topic2SQL(self, topic):
        """
        Generic way to generate a table name from MQTT topic
        :param topic: MQTT topic
        :return: SQL confirm table name
        """
        return topic

    # @classmethod
    # def from_mapper_type(cls, parent, _mapper_type):
    #     """
    #     Construct a structure from structure type
    #     """
    #     return BaseMapper(parent)  # ToDo: Factory of mappers

    async def on_save_session(self, session):
        """
        Persist the session
        :param session: The current session
        """

    async def on_find_session(self, client_id):
        """
        Retrieve a session by client ID
        :param client_id:
        :returns The session
        """

    async def on_del_session(self, client_id):
        """
        Delete a session by ClientID
        :param client_id:
        """

    async def on_broker_post_shutdown(self):
        """
        Shutdown the database if broker has shut down
        """





# Handlers
ON_SAVE_SESSION = 'on_save_session'
ON_FIND_SESSION = 'on_find_session'
ON_DEL_SESSION = 'on_del_session'
ON_BROKER_POST_SHUTDOWN = 'on_broker_post_shutdown'
ON_MQTT_PACKET_RECEIVED = 'on_mqtt_packet_received'

# YAML
BASE_CONFIG_PATH = 'plugins'
PAYLOAD_CONFIG = 'payload'

AMQTT_PLUGIN_NAME = 'amqtt_db'

TOPIC_ENGINE = 'topic_engine'
DB_CONNECT_STRING = 'db_connection'

DB_STRUCTURE = 'db_structure'




class AMQTTDBException(Exception):
    """
    Base Exception for module
    """


class TopicNotFound(AMQTTDBException):
    """
    Raised if a topic is not found
    """


class TypeNotFound(AMQTTDBException):
    """
    Raised if a type is not found
    """


class NoSAMappingForType(AMQTTDBException):
    """
    Raised if no SA mapping for a given python type is found
    """


class NoPayloadDefinition(AMQTTDBException):
    """
    Raised if no payload definition is found in the config
    """


class WrongPayloadDefinition(AMQTTDBException):
    """
    Raised if the payload config cannot be loaded by the topic engine
    """



from sqlalchemy import Table
from sqlalchemy.ext.declarative import declarative_base


class Base(object):
    """
    Give the SA-Tables an attribute decl_class which references the declarative class.
    """
    @classmethod
    def __table_cls__(cls, *args, **kwargs):
        t = Table(*args, **kwargs)
        t.decl_class = cls
        return t


Base = declarative_base(cls=Base)
